/*
** memory.c - Example demonstrating host access to memory banks
**            on an ADM-X* series card.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XP
**   o ADM-XPL
**   o ADP-DRC-II
**   o ADP-WRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**
** (C) Copyright 2005-2007 Alpha Data
*/

//#define ZAP

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>

#include <admxrc2.h>

#include "../common/common.h"

typedef int BOOL;

#define min(a, b) ((a) < (b)) ? (a) : (b)

/*
** Register indices and bit definitions
*/

#define BANK_REG				    	(0x0)				/* Bank select register */
#define BANK_REG_BANKMASK		    	(0xfU << 0)

#define PAGE_REG				    	(0x1)				/* Page register */
#define PAGE_REG_PAGEMASK		    	(0x7ffU << 0)

#define MEMCTL_REG				    	(0x2)				/* Memory control register */
#define MEMCTL_REG_MEMRST               (0x1U << 0)

#define STATUS_REG				    	(0x4)				/* Status register */
#define STATUS_REG_LCLKLOCKED	    	(0x1U << 0)
#define STATUS_REG_SHIFT_DCMLOCKED     	(8)

#define MODEx_REG(n)			    	(0x10 + (n))		/* Bank 'n' mode register */
#define MODEx_REG_DDR2SSRAM_BLMASK  	(0x3U << 0)			/* DDR2 SSRAM burst length mask */
#define MODEx_REG_DDR2SSRAM_BL2	    	(0x0U << 0)			/* DDR2 SSRAM burst length 2 */
#define MODEx_REG_DDR2SSRAM_BL4	    	(0x1U << 0)			/* DDR2 SSRAM burst length 4 */
#define MODEx_REG_DDR2SSRAM_DLLON   	(0x0U << 2)			/* DDR2 SSRAM DLL enable */
#define MODEx_REG_DDR2SSRAM_DLLOFF  	(0x1U << 2)			/* DDR2 SSRAM DLL disable */

#define MODEx_REG_ZBTSSRAM_FLOWTHRU	    (0x0U << 0)			/* ZBT SSRAM pipelined mode disable */
#define MODEx_REG_ZBTSSRAM_PIPELINE	    (0x1U << 0)			/* ZBT SSRAM pipelined mode enable */

#define MODEx_REG_DDRSDRAM_UNBUFFERED   (0x0U << 0)
#define MODEx_REG_DDRSDRAM_REGISTERED   (0x1U << 0)
#define MODEx_REG_DDRSDRAM_X8X16        (0x0U << 1)
#define MODEx_REG_DDRSDRAM_X4           (0x1U << 1)
#define MODEx_REG_DDRSDRAM_ROWBITS_12   (0x0U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_13   (0x1U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_14   (0x2U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_15   (0x3U << 2)
#define MODEx_REG_DDRSDRAM_COLBITS_M4   (0x0U << 4)         /* col bits = row bits - 4 */
#define MODEx_REG_DDRSDRAM_COLBITS_M3   (0x1U << 4)         /* col bits = row bits - 3 */
#define MODEx_REG_DDRSDRAM_COLBITS_M2   (0x2U << 4)         /* col bits = row bits - 2 */
#define MODEx_REG_DDRSDRAM_COLBITS_M1   (0x3U << 4)         /* col bits = row bits - 1 */
#define MODEx_REG_DDRSDRAM_BANKBITS_0   (0x0U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_1   (0x1U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_2   (0x2U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_3   (0x3U << 6)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_1  (0x0U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_2  (0x1U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_4  (0x2U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_8  (0x3U << 8)

#define SPD_REG                         (0x5)
#define SPD_REG_DO                      (0x1U << 31)
#define SPD_REG_BUSY                    (0x1U << 31)
#define SPD_REG_SHIFT_ADDR              (0)
#define SPD_REG_SHIFT_WRDATA            (8)
#define SPD_REG_SHIFT_RDDATA            (16)
#define SPD_REG_ERR                     (0x1U << 24)
#define SPD_REG_SHIFT_SLOT              (25)

#if ZAP
#define TUNECTL_REG                     (0x6)
#define TUNECTL_REG_SHIFT_ADDR          (0)
#define TUNECTL_REG_MASK_ADDR           (0xffU << TUNECTL_REG_SHIFT_ADDR)
#define TUNECTL_REG_GO                  (0x1U << 8)

#endif /* ZAP */
#if ZAP
#define TUNEDAT_REG                     (0x7)

#endif /* ZAP */

#define ERRATA_REG				        (0x8)				/* Errata register */
#define ERRATA_REG_220MHZREFCLK         (0x1U << 0)

/*
** Random number generator using Mersenne Twister
*/

#define MERS_N (351)
#define MERS_M (175)
#define MERS_R (19)
#define MERS_U (11)
#define MERS_S (7)
#define MERS_T (15)
#define MERS_L (17)
#define MERS_A (0xE4BD75F5)
#define MERS_B (0x655E5280)
#define MERS_C (0xFFD58000)

struct _MTState {
  uint32_t mt[MERS_N]; /* state vector */
  int mti; /* index into mt */
};
typedef struct _MTState MTState;

const unsigned long		memWindowAddr = 0x200000U;	 /* In bytes */
const unsigned long		pageSize  = 0x200000U;   /* In bytes */
const unsigned long		pageShift = 21;

const unsigned long		maxReportedErr = 10;

/* Chunk size for generating pattern data to write to memory banks */
#define GENERATE_CHUNK (0x1U << 20)

/* Max. number of memory banks supported by this application */
#define MAX_BANK (16)

/* Max. logical width of a memory bank, in bytes */
#define MAX_BANK_WIDTH (64)

struct _Globals {
  uint8_t revLookup[0x100];
  struct timeval lastSpin;
  MTState mtState;
  volatile void* pFpgaSpace;
  volatile uint32_t* pFpgaRegs;
  uint8_t* pMemoryWindow;
  uint8_t* pBuf5a;
  uint8_t* pBufOwnAddr;
  uint8_t* pBufRevAddr;
  uint8_t* pBufRandom;
  uint8_t* pBufChk;
  ADMXRC2_HANDLE hCard;
  uint32_t dmaMode;
  uint32_t bankTestMask;
  BOOL bBankDdr[MAX_BANK];
  unsigned int bankWidth[MAX_BANK];
  uint64_t bankSize[MAX_BANK];
  BOOL bUse64Bit;
};

typedef struct _Globals Globals;

void
seed(
  MTState* state,
  uint32_t seed)
{
  /* re-seed generator */
  state->mt[0] = seed;
  for (state->mti = 1; state->mti < MERS_N; state->mti++) {
    state->mt[state->mti] = (1812433253UL * (state->mt[state->mti - 1] ^ (state->mt[state->mti - 1] >> 30)) + state->mti);
  }
}

static uint32_t
random32(
  MTState* state)
{
  uint32_t y;
  
  if (state->mti >= MERS_N) {
    /* generate MERS_N words at one time */
    const uint32_t LOWER_MASK = (1LU << MERS_R) - 1; /* lower MERS_R bits */
    const uint32_t UPPER_MASK = -1L  << MERS_R;      /* upper (32 - MERS_R) bits */
    static const uint32_t mag01[2] = { 0, MERS_A };
    
    int kk;
    for (kk = 0; kk < (MERS_N - MERS_M); kk++) {
      y = (state->mt[kk] & UPPER_MASK) | (state->mt[kk + 1] & LOWER_MASK);
      state->mt[kk] = state->mt[kk + MERS_M] ^ (y >> 1) ^ mag01[y & 1];
    }
    
    for (; kk < MERS_N-1; kk++) {    
      y = (state->mt[kk] & UPPER_MASK) | (state->mt[kk + 1] & LOWER_MASK);
      state->mt[kk] = state->mt[kk + (MERS_M - MERS_N)] ^ (y >> 1) ^ mag01[y & 1];}      
    
    y = (state->mt[MERS_N - 1] & UPPER_MASK) | (state->mt[0] & LOWER_MASK);
    state->mt[MERS_N - 1] = state->mt[MERS_M - 1] ^ (y >> 1) ^ mag01[y & 1];
    state->mti = 0;
  }
  y = state->mt[state->mti++];
  
  /* Tempering (may be omitted) */
  y ^=  y >> MERS_U;
  y ^= (y << MERS_S) & MERS_B;
  y ^= (y << MERS_T) & MERS_C;
  y ^=  y >> MERS_L;
  
  return y;
}

static ADMXRC2_STATUS
reverse(
  Globals* pGlob,
  uint32_t v)
{
  uint8_t a, b, c, d;
  
  a = pGlob->revLookup[(v >> 0) & 0xff];
  b = pGlob->revLookup[(v >> 8) & 0xff];
  c = pGlob->revLookup[(v >> 16) & 0xff];
  d = pGlob->revLookup[(v >> 24) & 0xff];
  return ((uint32_t)a << 24) | ((uint32_t)b << 16) | ((uint32_t)c << 8) | ((uint32_t)d << 0);
}

static void
spin(
  Globals* pGlob,
  unsigned int* spindex)
{
  const char spinner[] = "-\\|/";
  const struct timeval interval = make_time(0, 100000);
  static BOOL bFirst = TRUE;
  struct timeval now;

  now = get_time();
  if (bFirst || time_after(&now, &pGlob->lastSpin)) {
    printf("%c\010", spinner[*spindex]);
    fflush(stdout);
    *spindex = (*spindex + 1) & 0x3U;
    if (bFirst) {
      pGlob->lastSpin = time_add(&now, &interval);
    } else {
      pGlob->lastSpin = time_add(&pGlob->lastSpin, &interval);
    }
    bFirst = FALSE;
  }
}

static void wordToString(
  char* str,
  uint8_t* buff,
  unsigned int n)
{
  unsigned int i;
  char s[8];
  
  strcpy(str, "0x");
  for (i = n; i > 0;) {   
    i--;
    sprintf(s, "%2.2lx", (unsigned long)buff[i]);
    strcat(str, s);
    if (i != 0) {
      strcat(str, "_");
    }
  }
}

static BOOL
verifyWord(
  uint8_t* buff,
  uint8_t* exp,
  uint8_t* act,
  uint8_t* mask,
  unsigned int width)
{
  unsigned int i;
  
  for (i = 0; i < width; i++) {
    act[i] = *buff++ & mask[i];
  }
  for (i = 0; i < width; i++) {
    if (act[i] != exp[i]) {
      return TRUE;
    }
  }
  return FALSE;
}

static void
displayError(
  uint64_t location,
  uint8_t* exp,
  uint8_t* act,
  unsigned int width,
  unsigned long nErr)
{
  char strExp[MAX_BANK_WIDTH * 3 + 2];
  char strAct[MAX_BANK_WIDTH * 3 + 2];
  
  if (nErr <= maxReportedErr) {
    wordToString(strExp, exp, width);
    wordToString(strAct, act, width);
    if (nErr == 1) {
      printf("\n");
    }
    printf("*** Loc 0x%1.1lx_%8.8lx: exp=%s act=%s\n",
	   (unsigned long)(location >> 32), (unsigned long)(uint32_t)location,
	   strExp, strAct);
  } else if (nErr == maxReportedErr + 1) {
    printf("*** [More errors]\n");
  }
}

static void
displayMask(
  unsigned int bankWidth,
  uint8_t* mask,
  char* str)
{
  unsigned int i;
  char* p = str;
  
  for (i = 0; i < bankWidth; i++) {
    p += sprintf(p, "%2.2lx", (unsigned long) mask[bankWidth - i - 1]);
    if (i != (bankWidth - 1)) {
      *p++ = '_';
    }
  }
  *p = '\0';
}

static ADMXRC2_STATUS
readMemory(
  Globals* pGlob,
  const void* buffer,
  uint64_t ramOffset,
  unsigned long	length,
  int useDMA)
{
  ADMXRC2_STATUS status;
  uint8_t* dst = (uint8_t*)buffer;
  uint8_t* src;
  unsigned long pgidx;
  unsigned long pgoffs;
  unsigned long chunk;
  
  while (length) {
    pgidx = (unsigned long)(ramOffset >> pageShift);
    pgoffs = (unsigned long)ramOffset & (pageSize - 1);
    
    chunk = pageSize - pgoffs;
    if (chunk > length) {
      chunk = length;
    }
    
    /* Set the page register */
    pGlob->pFpgaRegs[PAGE_REG] = pgidx & PAGE_REG_PAGEMASK;
    pGlob->pFpgaRegs[PAGE_REG]; /* make sure it's got there */
    
    if (useDMA) {
      status = ADMXRC2_DoDMAImmediate(pGlob->hCard,
				      dst,
				      chunk,
				      memWindowAddr + pgoffs,
				      ADMXRC2_LOCALTOPCI,
				      ADMXRC2_DMACHAN_ANY,
				      pGlob->dmaMode,
				      0,
				      NULL,
				      NULL);
      if (status != ADMXRC2_SUCCESS) {
	return status;
      }
    } else {
      src = pGlob->pMemoryWindow + pgoffs;
      memcpy(dst, src, chunk);
    }
    dst += chunk;
    ramOffset += chunk;
    length -= chunk;
  }
  
  return ADMXRC2_SUCCESS;
}

static ADMXRC2_STATUS
writeMemory(
  Globals* pGlob,
  const void* buffer,
  uint64_t ramOffset,
  unsigned long	length,
  int useDMA)
{
  ADMXRC2_STATUS status;
  uint8_t* dst;
  uint8_t* src = (uint8_t*)buffer;
  unsigned long pgidx;
  unsigned long pgoffs;
  unsigned long chunk;
  
  while (length) {
    pgidx = (unsigned long)(ramOffset >> pageShift);
    pgoffs = (unsigned long)ramOffset & (pageSize - 1);
    
    chunk = pageSize - pgoffs;
    if (chunk > length) {
      chunk = length;
    }
    
    /* Set the page register */
    pGlob->pFpgaRegs[PAGE_REG] = pgidx & PAGE_REG_PAGEMASK;
    pGlob->pFpgaRegs[PAGE_REG]; /* make sure it's got there */
    
    if (useDMA) {
      status = ADMXRC2_DoDMAImmediate(pGlob->hCard,
				      src,
				      chunk,
				      memWindowAddr + pgoffs,
				      ADMXRC2_PCITOLOCAL,
				      ADMXRC2_DMACHAN_ANY,
				      pGlob->dmaMode,
				      0,
				      NULL,
				      NULL);
      if (status != ADMXRC2_SUCCESS) {
	return status;
      }
    } else {
      dst = pGlob->pMemoryWindow + pgoffs;
      memcpy(dst, src, chunk);
    }
    src += chunk;
    ramOffset += chunk;
    length -= chunk;
  }
  
  return ADMXRC2_SUCCESS;
}

static void
maskExpectedWord(
  uint8_t* exp,
  uint8_t* data,
  uint8_t* mask,
  unsigned int width)
{
  unsigned int i;
  
  for (i = 0; i < width; i++) {
    exp[i] = data[i] & mask[i];
  }
}

static void
generate5A(
  uint8_t* buffer,
  unsigned long size,
  unsigned int width,
  int bIsDdr)
{
  unsigned long i;
  unsigned int j;
  int x = 0;
  uint8_t patt;
  assert(width >= 2);
  
  if (bIsDdr) {
    unsigned int halfwidth = width >> 1;
    
    assert((size % (width << 1)) == 0);
    for (i = 0; i < size;) {
      patt = x ? 0xAA : 0x55;
      for (j = 0; j < halfwidth; j++) {
	buffer[i++] = patt;
      }
      patt ^= 0xff;
      for (j = 0; j < halfwidth; j++) {
	buffer[i++] = patt;
      }
      x ^= 0x1;
    }
  } else {
    assert((size % (width << 2)) == 0);
    for (i = 0; i < size;) {
      patt = x ? 0xAA : 0x55;
      for (j = 0; j < width; j++) {
	buffer[i++] = patt;
      }
      patt ^= 0xff;
      for (j = 0; j < width; j++) {
	buffer[i++] = patt;
      }
      x ^= 0x1;
    }
  }
}

static void
generateOwnAddr(
  uint32_t* buffer,
  uint32_t offset,
  unsigned long size)
{
  unsigned long i;
  
  assert((offset & 0x3U) == 0);
  assert((size & 0x3U) == 0);
  
  size >>= 2;
  for (i = 0; i < size; i++) {
    *buffer++ = offset++;
  }
}

static void
generateRevAddr(
  Globals* pGlob,
  uint32_t* buffer,
  unsigned long offset,
  unsigned long size)
{
  unsigned long i;
  
  assert((offset & 0x3U) == 0);
  assert((size & 0x3U) == 0);
  
  size >>= 2;
  for (i = 0; i < size; i++) {
    *buffer++ = reverse(pGlob, offset);
    offset++;
  }
}

static void
generateRandom(
  MTState* mtState,
  uint32_t* buffer,
  unsigned long size)
{
  unsigned long i;
  
  assert((size & 0x3U) == 0);
  
  size >>= 2;
  for (i = 0; i < size; i++) {
    *buffer++ = random32(mtState);
  }
}

/*
** Check that local bus clock DCM has achieved lock.
*/
static BOOL
waitForLclkDcm(
  Globals* pGlob)
{
  uint32_t status;
  
  /*
  ** Delay for 500ms to allow DLLs/DCMs time to lock, and then verify
  ** that they are locked as expected.
  */
  sleep_ms(500);
  status = pGlob->pFpgaRegs[STATUS_REG];
  printf("STATUS=0x%8.8lx\n", (unsigned long)status);
  if (!(status & STATUS_REG_LCLKLOCKED)) {
    printf("*** LCLK DCM not locked\n");
    return FALSE;
  }
  return TRUE;
}

/*
** Check that DCMs/PLLs in memory clock domain have achieved lock, and that
** any memory banks that are self-training have trained themselves without
** error.
*/
static BOOL
checkForMemoryLock(
  Globals* pGlob,
  unsigned int numLockFlags,
  BOOL bVerifyTraining)
{
  uint32_t status, mask, val32;
  
  mask = (1 << numLockFlags) - 1;
  status = pGlob->pFpgaRegs[STATUS_REG];
  printf("STATUS=0x%8.8lx\n", (unsigned long)status);
  val32 = status >> STATUS_REG_SHIFT_DCMLOCKED;
  val32 &= mask;
  if (val32 != mask) {
    printf("*** Not all memory lock flags asserted: expected=0x%lx actual=0x%lx\n",
	   (unsigned long)mask, (unsigned long)val32);
    return FALSE;
  }
  if (bVerifyTraining) {
    val32 = status >> (STATUS_REG_SHIFT_DCMLOCKED + numLockFlags);
    val32 &= pGlob->bankTestMask;
    if (val32 != pGlob->bankTestMask) {
      printf("*** Not all memory banks trained: expected=0x%lx actual=0x%lx\n",
	     (unsigned long)pGlob->bankTestMask, (unsigned long)val32);
      return FALSE;
    }
  }
  
  return TRUE;
}

#if ZAP
static BOOL
getDimmMode(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
  unsigned int bank,
  uint32_t* pMode,
  BOOL* x4)
{
  uint8_t spd[128];
  BOOL bPresent;
  unsigned long size;
  unsigned int rows, cols, banks, pbanks;
  unsigned int devWidth;
  uint8_t clMask;
  uint8_t modAttr;
  BOOL bRegistered;
  uint32_t mode = 0;
  
  bPresent = readSpd(pGlob, bank, spd);
  if (bPresent) {
    rows = spd[3];      /* Number of bits used for row address on devices */
    cols = spd[4];      /* Number of bits used for column address on devices */
    pbanks = spd[5];    /* Number of physical banks on DIMM */
    devWidth = spd[13]; /* DQ width of devices */
    banks = spd[17];    /* Number of banks on devices */
    clMask = spd[18];   /* CAS latencies supported by devices */
    modAttr = spd[21];  /* DIMM module attributes */
    switch (devWidth) {
    case 4:
    case 8:
    case 16:
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): unknown device data width (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)devWidth);
      return FALSE;
    }
    if (banks < 1 || banks > 8) {
      printf(" *** Bank %ld (DDR SDRAM): unsupported device banks value (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    if (!(clMask & 0x8U)) {
      printf(" +++ Bank %ld (DDR SDRAM): CAS latency 2.5 not supported by devices, not testing\n",
	     (unsigned long)bank);
      return FALSE;
    }
    switch (modAttr) {
    case 0x20: /* Unbuffered */
      bRegistered = FALSE;
      break;
    case 0x26: /* Registered */
      bRegistered = TRUE;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid DIMM module attributes value (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)modAttr);
      return FALSE;
    }
    switch (rows) {
    case 12:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_12;
      break;
    case 13:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
      break;
    case 14:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_14;
      break;
    case 15:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_15;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of row bits (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)rows);
      return FALSE;
    }
    switch (cols + 4 - rows) {
    case 0:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
      break;
    case 1:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
      break;
    case 3:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M1;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of column bits (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)cols);
      return FALSE;
    }
    switch (banks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_0;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_1;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of banks (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    switch (pbanks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_2;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_4;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_8;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of physical banks (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)pbanks);
      return FALSE;
    }
    if (bRegistered) {
      mode |= 0x1U;
    }
    if (devWidth == 4) {
      mode |= 0x2U;
    }
    pBankInfo[bank].Size = (1 << (rows + cols - 1)) * banks * pbanks;
    size = (pBankInfo[bank].Size >> 20) * (pBankInfo[bank].Width >> 3);
    printf(" Bank %ld (DDR SDRAM): x%ld %s rows=%ld cols=%ld banks=%ld physBanks=%ld %ldMB\n",
	   (unsigned long)bank,
	   (unsigned long)devWidth,
	   bRegistered ? "registered" : "unbuffered",
	   (unsigned long)rows, (unsigned long)cols,
	   (unsigned long)banks, (unsigned long)pbanks, (unsigned long)size);
  } else {
    printf(" +++ Bank %ld (DDR SDRAM): SPD data not found, not testing\n",
	   (unsigned long)bank);
    return FALSE;
  }
  
  /* Bank SPD seems OK */
  *pMode = mode;
  *x4 = (devWidth == 4);
  return TRUE;
}

#endif /* ZAP */
static void
setMemoryPortConfiguration(
  Globals* pGlob,
  uint32_t* modes)
{
  unsigned int i;
  
  printf("Configuring memory ports...\n");
  for (i = 0; i < MAX_BANK; i++) {
    if (!(pGlob->bankTestMask & (0x1UL << i))) {
      continue;
    }
    pGlob->pFpgaRegs[MODEx_REG(i)] = modes[i];
  }
}

static void
scanMemoryPorts(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
#if ZAP
  char** reloadBit,
#endif /* ZAP */
  uint32_t* modes)
{
  unsigned int i;
#if ZAP
  BOOL dimmX4[4];
#endif /* ZAP */
  
  /* Initialize memory bank config registers where necessary */
  printf("Scanning memory ports...\n");
  for (i = 0; i < pCardInfo->NumRAMBank; i++) {
    if (!(pGlob->bankTestMask & (0x1U << i))) {
      continue;
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SRAM_DDR2) {
      /* Select BL4 for DDR-II SSRAM and ensure DLL is enabled */
      modes[i] = MODEx_REG_DDR2SSRAM_BL4 | MODEx_REG_DDR2SSRAM_DLLON;
      pGlob->bBankDdr[i] = TRUE;
      printf(" Bank %ld (DDR-II SRAM): burst length 4, DLL enabled\n",
	     (unsigned long) i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTP) {
      /* Select pipelined mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_PIPELINE;
      pGlob->bBankDdr[i] = FALSE;
      printf(" Bank %ld (ZBT SSRAM): pipelined mode\n",
	     (unsigned long) i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTFT) {
      /* Select flowthrough mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_FLOWTHRU;
      pGlob->bBankDdr[i] = FALSE;
      printf(" Bank %ld (ZBT SSRAM): flowthrough mode\n",
	     (unsigned long) i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR) {
      uint32_t mode;
      
      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXPL:
      case ADMXRC2_BOARD_ADMXP:
	mode = 0;
	if (0x800000 == pBankInfo[i].Size) {
	  /* unbufferred, row bits 13, col bits 9, bank bits 2, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR SDRAM): unbuffered, row 13, col 9, bank 2, phys banks 1\n",
		 (unsigned long) i);
	} else if (0x1000000 == pBankInfo[i].Size) {
	  /* unregistered, row bits 13, col bits 10, bank bits 2, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR SDRAM): unbuffered, row 13, col 10, bank 2, phys banks 1\n",
		 (unsigned long) i);
	} else if (0x2000000 == pBankInfo[i].Size) {
	  /* unregistered, row bits 13, col bits 11, bank bits 2, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR SDRAM): unbuffered, row 13, col  11, bank 2, phys banks 1\n",
		 (unsigned long) i);
	} else {
	  pGlob->bankTestMask &= ~(0x1U << i);
	  printf(" +++ Bank %ld (DDR SDRAM): invalid size (0x%lx words), not testing\n",
		 (unsigned long)i, (unsigned long)pBankInfo[i].Size);
	}
	modes[i] = mode;
	pGlob->bBankDdr[i] = TRUE;
	break;
      default:
	/* Should never get here */
	assert(FALSE);
      }
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR2) {
      uint32_t mode;
      
      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXRC4FX:
      case ADMXRC2_BOARD_ADMXRC5LX:
      case ADMXRC2_BOARD_ADMXRC5T1:
	mode = 0;
	if (0x2000000 == pBankInfo[i].Size) {
	  /* unbufferred, row bits 13, col bits 9, bank bits 2, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR-II SDRAM): unbuffered, row 13, col 10, bank 3, phys banks 1\n",
		 (unsigned long) i);
	} else {
	  pGlob->bankTestMask &= ~(0x1U << i);
	  printf(" +++ Bank %ld (DDR-II SDRAM): invalid size (0x%lx words), not testing\n",
		 (unsigned long)i, (unsigned long)pBankInfo[i].Size);
	}
	modes[i] = mode;
	pGlob->bBankDdr[i] = TRUE;
	break;
      default:
	/* Should never get here */
	assert(FALSE);
      }
    }
  }
}

int
main(int argc, char *argv[])
{
  Globals glob;
  int ret = 0;
  ADMXRC2_STATUS status;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  ADMXRC2_BANK_INFO bankInfo[MAX_BANK];
  const char* filename;
  char* design;
  Option options[14];
  Arguments arguments;
  unsigned int i, rep;
  unsigned int numLockFlags;
  int dma;
  uint64_t memorySize;
  BOOL bUseDma;
  BOOL bUsePio;
  BOOL bMeasurePerf;
  BOOL bSetMclk, bAllow64Bit;
  BOOL bUseModule;
  uint32_t modes[MAX_BANK];
  char* moduleName;
  double lclkFreq, actualLclkFreq;
  double mclkFreq, actualMclkFreq;
  unsigned long nrep;
  BOOL bFailed;
  uint8_t mask[MAX_BANK_WIDTH];
  BOOL bUseMask;
  BOOL b220MHzRefClk;
  BOOL bRetryReads;
  BOOL bVerifyTraining;

  /***********************************************************************
  ** 1. Initialize stuff so that we know where we are if we have to     **
  **    clean up if some later step fails.                              **
  ***********************************************************************/

  glob.lastSpin.tv_sec = 0;
  glob.lastSpin.tv_usec = 0;
  for (i = 0; i < 0x100; i++) {
    unsigned int j;
    uint8_t val;
    
    val = 0;
    for (j = 0; j < 8; j++) {
      val |= ((i >> j) & 0x1U) << (7 - j);
    }
    glob.revLookup[i] = val;
  }
  glob.pBuf5a = NULL;
  glob.pBufOwnAddr = NULL;
  glob.pBufRevAddr = NULL;
  glob.pBufRandom = NULL;
  glob.pBufChk = NULL;
  glob.hCard = ADMXRC2_HANDLE_INVALID_VALUE;
  
  /***********************************************************************
  ** 2. Set up the data structure that defines command-line syntax.     **
  ***********************************************************************/

  options[0].key = "banks";
  options[0].type = Option_hex;
  options[0].def.hexVal = 0xffffU;
  options[0].meaning = "bitmask of banks to test";
  options[1].key = "card";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "ID of card to use";
  options[2].key = "index";
  options[2].type = Option_uint;
  options[2].def.uintVal = 0;
  options[2].meaning = "index of card to use";
  options[3].key = "repeat";
  options[3].type = Option_uint;
  options[3].def.uintVal = 1;
  options[3].meaning = "number of repetitions";
  options[4].key = "perf";
  options[4].type = Option_boolean;
  options[4].def.booleanVal = TRUE;
  options[4].meaning = "TRUE => measure DMA-to-bank performance";
  options[5].key = "usedma";
  options[5].type = Option_boolean;
  options[5].def.booleanVal = TRUE;
  options[5].meaning = "TRUE => test using DMA";
  options[6].key = "usepio";
  options[6].type = Option_boolean;
  options[6].def.booleanVal = FALSE;
  options[6].meaning = "TRUE => test using programmed I/O";
  options[7].key = "64";
  options[7].type = Option_boolean;
  options[7].def.booleanVal = FALSE;
  options[7].meaning = "TRUE => use 64 bit local bus";
  options[8].key = "lclk";
  options[8].type = Option_float;
  options[8].def.floatVal = 33.0;
  options[8].meaning = "local bus clock frequency, in MHz";
  options[9].key = "mclk";
  options[9].type = Option_float;
  options[9].def.floatVal = 33.0;
  options[9].meaning = "MCLK frequency, in MHz";
  options[10].key = "module";
  options[10].type = Option_string;
  options[10].def.stringVal = "<none>";
  options[10].meaning = "target specified XRM module";
  options[11].key = "mask";
  options[11].type = Option_string;
  options[11].def.stringVal = "<none>";
  options[11].meaning = "physical bit lanes to test";
  options[12].key = "refclk220";
  options[12].type = Option_boolean;
  options[12].def.booleanVal = FALSE;
  options[12].meaning = "TRUE => use 220MHz IDELAY refclk (V5 only)";
  options[13].key = "retry";
  options[13].type = Option_boolean;
  options[13].def.booleanVal = FALSE;
  options[13].meaning = "TRUE => retry reads on verification error";
  arguments.nOption = 14;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;
	
  /***********************************************************************
  ** 3. Parse the command line                                          **
  ***********************************************************************/

  parse_command_line("memory",
		     argc,
		     argv,
		     &arguments);
  if (options[1].specified && options[2].specified) {
    printf("*** Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }
  
  nrep = options[3].value.uintVal;
  bMeasurePerf = options[4].value.booleanVal;
  bUseDma = options[5].value.booleanVal;
  bUsePio = options[6].value.booleanVal;
  glob.bUse64Bit = options[7].value.booleanVal;
  moduleName = options[10].value.stringVal;
  bUseModule = strcmp(moduleName, "<none>") != 0;
  b220MHzRefClk = options[12].value.booleanVal;
  bRetryReads = options[13].value.booleanVal;
  
  /***********************************************************************
  ** 4. Open the selected card                                          **
  ***********************************************************************/

  if (!options[2].specified) {
    ADMXRC2_CARDID cardID = options[1].value.uintVal;
    
    status = ADMXRC2_OpenCard(cardID, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[2].value.uintVal;
    
    status = ADMXRC2_OpenCardByIndex(index, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }
  
  /***********************************************************************
  ** 5. Get card information for the selected card. This will be used   **
  **    later to determine default clock frequencies and other model-   **
  **    specific settings.                                              **
  ***********************************************************************/

  status = ADMXRC2_GetCardInfo(glob.hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    return -1;
  }

  /***********************************************************************
  ** 6. Figure out model-specific settings from card type.              **
  ***********************************************************************/

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 33.33;
    mclkFreq = 40.0;
    numLockFlags = 2;
    bVerifyTraining = FALSE;
    break;
    
  case ADMXRC2_BOARD_ADMXRC_P:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 33.33;
    mclkFreq = 40.0;
    numLockFlags = 1;
    bVerifyTraining = FALSE;
    break;
    
  case ADMXRC2_BOARD_ADMXRC2_LITE:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 40.0;
    mclkFreq = 66.67;
    numLockFlags = 3;
    bVerifyTraining = FALSE;
    break;
    
  case ADMXRC2_BOARD_ADMXRC2:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 50.0;
    mclkFreq = 66.67;
    numLockFlags = 3;
    bVerifyTraining = FALSE;
    break;
    
  case ADMXRC2_BOARD_ADMXPL:
    bAllow64Bit = TRUE;
    bSetMclk = FALSE;
    lclkFreq = 66.67;
    mclkFreq = 2.0 * lclkFreq;
    numLockFlags = 2;
    bVerifyTraining = FALSE;
    break;

    case ADMXRC2_BOARD_ADMXP:
      bAllow64Bit = TRUE;
      bSetMclk = FALSE;
      lclkFreq = 66.67;
      mclkFreq = 2.0 * lclkFreq;
      numLockFlags = 1;
      bVerifyTraining = FALSE;
      break;

  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 66.0;
    mclkFreq = 150.0; /* 150MHz is max 1X output frequency for DCM in low frequency mode */
    numLockFlags = 2;
    bVerifyTraining = FALSE;
    break;
    
  case ADMXRC2_BOARD_ADMXRC4FX:
    bAllow64Bit = TRUE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 200.0MHz MCLK => 100.0MHz in memory clock domain, 200.0MHz DDR-II SDRAM clock */
    mclkFreq = 200.0;
    numLockFlags = 2;
    bVerifyTraining = TRUE;
    break;

  case ADMXRC2_BOARD_ADMXRC5LX:
    bAllow64Bit = TRUE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
    mclkFreq = 266.67;
    numLockFlags = 3;
    bVerifyTraining = TRUE;
    break;

  case ADMXRC2_BOARD_ADMXRC5T1:
    bAllow64Bit = TRUE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
    mclkFreq = 266.67;
    numLockFlags = 3;
    bVerifyTraining = TRUE;
    break;

  default:
    printf("*** This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II-Lite\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    ret = -1;
    goto done;
  }
  
  if (glob.bUse64Bit && !bAllow64Bit) {
    printf("*** 64-bit local bus is not supported on the specified card.\n");
    ret = -1;
    goto done;
  }
  
  /***********************************************************************
  ** 7. Get the addresses for the FPGA registers and the memory access  **
  **    window, in our virtual address space.                           **
  ***********************************************************************/

  status = ADMXRC2_GetSpaceInfo(glob.hCard, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get space 0 info: %s\n",
	   ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  glob.pFpgaSpace = spInfo.VirtualBase;
  if (glob.pFpgaSpace == NULL) {
    printf("*** Failed to retrieve pointer to FPGA space\n");
    ret = -1;
    goto done;
  }
  glob.pFpgaRegs = (volatile uint32_t*)glob.pFpgaSpace;
  glob.pMemoryWindow = (uint8_t*)glob.pFpgaSpace;
  glob.pMemoryWindow += memWindowAddr;
  
  /***********************************************************************
  ** 8. Get memory bank information.                                    **
  **                                                                    **
  **    To determine the byte width of a memory bank, we first divide   **
  **    by 9. If there is no remainder, then that bank has parity bits, **
  **    and we have calculated the byte width. If there is a remainder, **
  **    then we assume the bank has no parity bits, in which case we    **
  **    should be able to calculate the byte width by dividing by 8.    **
  **                                                                    **
  **    For the ADM-XRC-II-Lite, we fudge the data a little. Each pair  **
  **    of 18-bit wide banks is combined to get a logical 32-bit wide   **
  **    bank.                                                           **
  ***********************************************************************/
  
  /* Get bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    if (cardInfo.RAMBanksFitted & (0x1UL << i)) {
      status = ADMXRC2_GetBankInfo(glob.hCard, i, &bankInfo[i]);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to get bank %ld info: %s\n", (unsigned long) i,
	       ADMXRC2_GetStatusString(status));
	return -1;
      }
    }
  }
  
  /*
  ** For the ADM-XRC-II-Lite, fix up the memory bank information to make
  ** it look like we have 32-bit wide banks (we combine two 18-bit physical
  ** banks into one logical bank).
  */
  if (cardInfo.BoardType == ADMXRC2_BOARD_ADMXRC2_LITE) {
    uint32_t banksFitted = 0x0U;
    
    cardInfo.NumRAMBank = 2;
    for (i = 0; i < cardInfo.NumRAMBank; i++) {
      unsigned int j = i * 2;
      BOOL bFittedA, bFittedB;
      BOOL bSame;
      
      bFittedA = (cardInfo.RAMBanksFitted >> j) != 0;
      bFittedB = (cardInfo.RAMBanksFitted >> (j + 1)) != 0;
      if (bFittedA && bFittedB) {
	bSame = (bankInfo[j].Width == bankInfo[j + 1].Width) &&
	  (bankInfo[j].Type == bankInfo[j + 1].Type);
	if (bSame) {
	  bankInfo[i].Size = min(bankInfo[j].Size, bankInfo[j + 1].Size);
	  bankInfo[i].Type = bankInfo[j].Type & bankInfo[j + 1].Type;
	  bankInfo[i].Fitted = bankInfo[j].Fitted & bankInfo[j + 1].Fitted;
	  banksFitted |= (0x1U << i);
	} else {
	  /* This logical bank has been deemed invalid */
	  printf("+++ Can't pair banks %ld and %ld due to different configuration\n",
		 (unsigned long)j, (unsigned long)(j + 1));
	}
      }
    }
    cardInfo.RAMBanksFitted = banksFitted;
  }
  
  /*
  ** For the ADM-XRC-5T1, pretend that bank 2 (DDR-II SSRAM) is not fitted.
  ** Support for this type of memory on V4/V5 is coming soon!
  */
  if (cardInfo.BoardType == ADMXRC2_BOARD_ADMXRC5T1) {
    cardInfo.RAMBanksFitted &= ~(0x1U << 2U);
  }
  
  /*
  ** Establish the bank test mask. This is the set of banks that are we
  ** have been instructed to test AND are fitted. This value may be
  ** further modified during the further stages of this application in
  ** order to remove banks that are deemed invalid.
  */
  glob.bankTestMask = options[0].value.hexVal & cardInfo.RAMBanksFitted;
  
  /* Process bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;
    
    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;
    
    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }
    
    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }
    
    if ((bankInfo[i].Width % 9) == 0) {
      /* Bank has parity bits - don't use to calculate bank size */
      glob.bankWidth[i] = bankInfo[i].Width / 9;
    } else {
      /* Bank doesn't have parity bits */
      glob.bankWidth[i] = bankInfo[i].Width / 8;
    }
  }
  
  /***********************************************************************
  ** 9. Parse any mask that might have been specified on the command-   **
  **    line. This may be useful for diagnosing certain types of memory **
  **    fault like a stuck data bit or a short to another data bit. The **
  **    mask specifies which bits of a memory word are actually         **
  **    verified during the memory tests.                               **
  ***********************************************************************/

  memset(mask, 0xffU, MAX_BANK_WIDTH);
  if (options[11].specified) {
    size_t i, n;
    uint8_t digit;
    unsigned int shift;
    char c;
    
    bUseMask = TRUE;
    n = strlen(options[11].value.stringVal);
    if (n < 1) {
      printf("*** Invalid value for -mask option: %s\n", options[11].value.stringVal);
      return -1;
    }
    memset(mask, 0x0U, MAX_BANK_WIDTH);
    for (i = 0; i < n; i++) {
      c = tolower(options[11].value.stringVal[i]);
      if (c >= 'a' && c <= 'f') {
	digit = c - 'a' + 10;
      } else if (c >= '0' && c <= '9') {
	digit = c - '0';
      } else {
	printf("*** Invalid digit %ld in value for -mask option: %s\n",
	       (unsigned long) (i + 1), options[11].value.stringVal);
	return -1;
      }
      shift = ((n - i - 1) & 0x1U) << 2;
      mask[(n - i - 1) >> 1] |= digit << shift;
    }
  } else {
    bUseMask = FALSE;
  }
  
  /***********************************************************************
  ** 10. Allocate the buffers that are used to hold test data patterns  **
  **     and the data read back during the memory tests.                **
  ***********************************************************************/

  glob.pBuf5a = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBuf5a == NULL) {
    printf("*** Failed to allocate 0x5A buffer - aborting\n");
    ret = -1;
    goto done;
  }
  
  glob.pBufOwnAddr = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufOwnAddr == NULL) {
    printf("*** Failed to allocate own address buffer - aborting\n");
    ret = -1;
    goto done;
  }
  
  glob.pBufRevAddr = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufRevAddr == NULL) {
    printf("*** Failed to allocate own address buffer - aborting\n");
    ret = -1;
    goto done;
  }
  
  glob.pBufRandom = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufRandom == NULL) {
    printf("*** Failed to allocate random data buffer - aborting\n");
    ret = -1;
    goto done;
  }
  
  glob.pBufChk = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufChk == NULL) {
    printf("*** Failed to allocate RAM readback buffer - aborting\n");
    ret = -1;
    goto done;
  }
  
  /***********************************************************************
  ** 11. Program the frequencies for LCLK and MCLK. The defaults are    **
  **     obtained from the board type, but may be overridden on the     **
  **     command-line.                                                  **
  ***********************************************************************/

  if (options[8].specified) {
    lclkFreq = options[8].value.floatVal;
  }
  
  if (options[9].specified) {
    mclkFreq = options[9].value.floatVal;
  }
  
  printf("Setting LCLK frequency to %.1fMHz\n", lclkFreq);
  status = ADMXRC2_SetClockRate(glob.hCard, ADMXRC2_CLOCK_LCLK, lclkFreq * 1.0e6, &actualLclkFreq);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to set LCLK to %.1fMHz: %s\n",
	   lclkFreq, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  
  if (bSetMclk) {
    printf("Setting MCLK frequency to %.1fMHz\n", mclkFreq);
    status = ADMXRC2_SetClockRate(glob.hCard, 1, mclkFreq * 1.0e6, &actualMclkFreq);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to set clock 1 to %.1fMHz: %s\n",
	     mclkFreq, ADMXRC2_GetStatusString(status));
      return -1;
    }
  }
  
  /***********************************************************************
  ** 12. Configure the FPGA with the "memory" bitstream.                **
  ***********************************************************************/

  design = glob.bUse64Bit ? "memory64" : "memory";
  filename = get_fpga_filename_dir(&cardInfo, "memory", design, &arguments);
  printf("Configuring FPGA with '%s'...\n", filename);
  status = ADMXRC2_ConfigureFromFile(glob.hCard, filename);
  if (status != ADMXRC2_SUCCESS) {
    printf ("*** Failed to load the bitstream '%s': %s\n",
	    filename, ADMXRC2_GetStatusString(status));
    return -1;
  }
  
  /***********************************************************************
  ** 13. After configuring the FPGA, it's necessary to wait a while     **
  **     to enable any DLLs, DCMs and PLLs that depend upon LCLK to     **
  **     achieve lock. The delay is a conservative 500ms, because we    **
  **     make no assumptions about whether or not an ES (engineering    **
  **     silicon) FPGA is being used.                                   **
  ***********************************************************************/

  if (!waitForLclkDcm(&glob)) {
    ret = -1;
    goto done;
  }
  
  /***********************************************************************
  ** 14. If running in 64-bit mode was requested on the command-line,   **
  **     configure Direct Slave space 0 for a 64-bit local bus. Also    **
  **     generate the DMA mode word for DMA using a 64-bit local bus.   **
  ***********************************************************************/

  if (glob.bUse64Bit) {
    uint32_t config = 0;
    
    config |= ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
    status = ADMXRC2_SetSpaceConfig(glob.hCard, 0, config);
    if (status != ADMXRC2_SUCCESS) {
      printf ("*** Failed to set space configuration: %s\n",
	      ADMXRC2_GetStatusString(status));
      return -1;
    }
  }
  
  glob.dmaMode = ADMXRC2_BuildDMAModeWord(cardInfo.BoardType,
					  glob.bUse64Bit ? ADMXRC2_IOWIDTH_64 : ADMXRC2_IOWIDTH_32,
					  0,
					  ADMXRC2_DMAMODE_USEREADY | ADMXRC2_DMAMODE_USEBTERM | ADMXRC2_DMAMODE_BURSTENABLE);
  
  /***********************************************************************
  ** 15. Here we use the memory bank information read in step 7 to      **
  **     determine the proper values for the MODE registers.            **
  ***********************************************************************/

  scanMemoryPorts(&cardInfo,
		  bankInfo,
		  &glob,
		  modes);

  /***********************************************************************
  ** 16. Establish the MODE registers, based on what was determined     **
  **     in step 14.                                                    **
  ***********************************************************************/

  setMemoryPortConfiguration(&glob, modes);

  /***********************************************************************
  ** 17. Specifying '+refclk220' on the command line enables the        **
  **     workaround for a problem with IDELAYCTRL reference clock on    **
  **     some very early engineering silicon Virtex-5 devices. This     **
  **     option should never be required, however.                      **
  ***********************************************************************/

  if (b220MHzRefClk) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC5LX:
    case ADMXRC2_BOARD_ADMXRC5T1:
      printf("Enabling V5 IDELAYCTRL reference clock workaround...\n");
      glob.pFpgaRegs[ERRATA_REG] = ERRATA_REG_220MHZREFCLK;
      glob.pFpgaRegs[ERRATA_REG];
      
      /*
      ** Because IDELAY reference clock frequency has changed, we need to
      ** reset the memory subsystem, which will recalibrate the IDELAYs.
      */
      printf("Resetting memory subsystem in order to recalibrate IDELAYs...\n");
      glob.pFpgaRegs[MEMCTL_REG] = 0x1U;
      glob.pFpgaRegs[MEMCTL_REG];
      sleep_ms(1);
      glob.pFpgaRegs[MEMCTL_REG] = 0x0U;
      glob.pFpgaRegs[MEMCTL_REG];
      sleep_ms(500);
      break;
      
    default:
      printf("+++ ignoring option +refclk220 (not running on Virtex-5)\n");
      break;
    }
  }
  
  /***********************************************************************
  ** 18. Reset the memory subsystem, and wait 500ms for DLLs, DCMs and  **
  **     PLLs to achieve lock, as well as allowing any self-training    **
  **     memory interfaces sufficient time to train themselves.         **
  ***********************************************************************/

  /* Check for memory DLL/DCM/PLL locking and training */
  if (!checkForMemoryLock(&glob, numLockFlags, bVerifyTraining)) {
    ret = -1;
    goto done;
  }
  
  /***********************************************************************
  ** 19. Figure out total size (in bytes) of memory banks that are to   **
  **     be tested.                                                     **
  ***********************************************************************/

  memorySize = 0;
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;
    
    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;
    
    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }
    
    if (!bankInfo[i].Fitted) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }
    
    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }
    
    glob.bankSize[i] = glob.bankWidth[i] * bankInfo[i].Size;
    memorySize += (glob.bankSize[i] >> 10);
  }
  
  /* If there's nothing to do, abort! */
  if (!glob.bankTestMask) {
    printf("*** Nothing to test - aborting.\n");
    ret = -1;
    goto done;
  }
  
  /***********************************************************************
  ** 20. This is where the memory tests begin. There are four different **
  **     tests:                                                         **
  **                                                                    **
  **      a) 55 and AA pattern, for detecting stuck / shorted data bits **
  **      b) Own address, for detecing faulty address lines             **
  **      c) Own address, bit reversed, for detecting faulty address    **
  **         lines                                                      **
  **      d) Randomly generated data, for detecting pattern-sensitive   **
  **         faults.                                                    **
  ***********************************************************************/
  
  printf("Total memory size to test : %ldkB (0x%lxkB)\n",
	 (unsigned long) memorySize, (unsigned long) memorySize);
  
  bFailed = FALSE;
  for (rep = 0; rep < nrep; rep++) {
    unsigned long nErr;
    unsigned int nRetry;
    unsigned long chunk;
    
    for (dma = 0; dma < 2; dma++) {
      unsigned int spindex = 0;
      uint64_t remaining, offset;
      uint64_t location;
      uint8_t exp[MAX_BANK_WIDTH];
      uint8_t act[MAX_BANK_WIDTH];
      unsigned int width;
      char maskStr[1024];
      
      if (!dma && !bUsePio) {
	continue;
      }
      if (dma && !bUseDma) {
	continue;
      }
      
      printf("Performing memory tests using %s\n", dma ? "DMA" : "programmed I/O");
      
      strcpy(maskStr, "         mask=");
      
      printf(" 5/A pattern...\n");
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
	if (!(glob.bankTestMask & (0x1U << i))) {
	  continue;
	}
	
	if (bUseMask) {
	  displayMask(glob.bankWidth[i], mask, maskStr + 14);
	} else {
	  maskStr[0] = '\0';
	}
	
	printf("  Bank %ld%s\r", (unsigned long)i, maskStr);
	glob.pFpgaRegs[BANK_REG] = i & BANK_REG_BANKMASK;
	glob.pFpgaRegs[BANK_REG];
	width = glob.bankWidth[i];
	generate5A((uint8_t*)glob.pBuf5a, GENERATE_CHUNK, width, glob.bBankDdr[i]);
        
	/* Write data to memory bank */
	remaining = glob.bankSize[i];
	offset = 0;
	while (remaining) {
	  chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  writeMemory(&glob, glob.pBuf5a, offset, chunk, dma != 0);
	  remaining -= chunk;
	  offset += chunk;
	  spin(&glob, &spindex);
	}
	
	/* Read back data from memory bank and check */
	remaining = glob.bankSize[i];
	offset = 0;
	location = 0;
	nErr = 0;
	nRetry = 20;
	chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	while (remaining) {
	  unsigned long pos;
	  uint64_t tempLoc = location;
	  uint64_t errLoc = 0; /* init. only to prevent warning from compiler */
	  unsigned long errPos = 0; /* init. only to prevent warning from compiler */
	  BOOL bOk = TRUE;
	  
	  readMemory(&glob, glob.pBufChk, offset, chunk, dma != 0);
	  tempLoc = location;
	  for (pos = 0; pos < chunk; pos += width) {
	    maskExpectedWord(exp, glob.pBuf5a + pos, mask, width);
	    if (verifyWord(glob.pBufChk + pos, exp, act, mask, width)) {
	      nErr++;
	      displayError(tempLoc, exp, act, width, nErr);
	      if (bOk) {
		bOk = FALSE;
		errLoc = tempLoc;
		errPos = pos;
	      }
	    }
	    tempLoc++;
	  }
	  if (bOk || !bRetryReads) {
	    remaining -= chunk;
	    offset += chunk;
	    location += chunk;
	    nRetry = 20;
	    chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  } else {
	    if (0 == nRetry) {
	      printf("[Giving up on location 0x%llx]\n", errLoc);
	      remaining -= chunk;
	      offset += chunk;
	      location += chunk;
	      nRetry = 20;
	      chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	    } else {
	      printf("[Retrying read at location 0x%llx]\n", errLoc);
	      remaining -= errPos;
	      offset += errPos;
	      location = errLoc;
	      nRetry--;
	      chunk = width;
	    }
	  }
	  spin(&glob, &spindex);
	}
	if (nErr) {
	  printf("*** %ld error(s) encountered while verifying data\n", (unsigned long) nErr);
	  bFailed = TRUE;
	  break;
	}
      }
      
      printf(" Own address...\n");
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
	if (!(glob.bankTestMask & (0x1U << i))) {
	  continue;
	}
	
	if (bUseMask) {
	  displayMask(glob.bankWidth[i], mask, maskStr + 14);
	} else {
	  maskStr[0] = '\0';
	}
	
	printf("  Bank %ld%s\r", (unsigned long)i, maskStr);
	glob.pFpgaRegs[BANK_REG] = i & BANK_REG_BANKMASK;
	glob.pFpgaRegs[BANK_REG];
	width = glob.bankWidth[i];
	
	/* Write data to memory bank */
	remaining = glob.bankSize[i];
	offset = 0;
	while (remaining) {
	  chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  generateOwnAddr((uint32_t*)glob.pBufOwnAddr, (uint32_t)(offset >> 2), chunk);
	  writeMemory(&glob, glob.pBufOwnAddr, offset, chunk, dma);
	  remaining -= chunk;
	  offset += chunk;
	  spin(&glob, &spindex);
	}
        
	/* Read back data from memory bank and check */
	remaining = glob.bankSize[i];
	offset = 0;
	location = 0;
	nErr = 0;
	nRetry = 20;
	chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	while (remaining) {
	  unsigned long pos;
	  uint64_t tempLoc = location;
	  uint64_t errLoc = 0; /* init. only to prevent warning from compiler */
	  unsigned long errPos = 0; /* init. only to prevent warning from compiler */
	  BOOL bOk = TRUE;
          
	  readMemory(&glob, glob.pBufChk, offset, chunk, dma);
	  generateOwnAddr((uint32_t*)glob.pBufOwnAddr, (uint32_t)(offset >> 2), chunk);
	  tempLoc = location;
	  for (pos = 0; pos < chunk; pos += width) {
	    maskExpectedWord(exp, glob.pBufOwnAddr + pos, mask, width);
	    if (verifyWord(glob.pBufChk + pos, exp, act, mask, width)) {
	      nErr++;
	      displayError(tempLoc, exp, act, width, nErr);
	      if (bOk) {
		bOk = FALSE;
		errLoc = tempLoc;
		errPos = pos;
	      }
	    }
	    tempLoc++;
	  }
	  if (bOk || !bRetryReads) {
	    remaining -= chunk;
	    offset += chunk;
	    location += chunk;
	    nRetry = 20;
	    chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  } else {
	    if (0 == nRetry) {
	      printf("[Giving up on location 0x%llx]\n", errLoc);
	      remaining -= chunk;
	      offset += chunk;
	      location += chunk;
	      nRetry = 20;
	      chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	    } else {
	      printf("[Retrying read at location 0x%llx]\n", errLoc);
	      remaining -= errPos;
	      offset += errPos;
	      location = errLoc;
	      nRetry--;
	      chunk = width;
	    }
	  }
	  spin(&glob, &spindex);
	}
	if (nErr) {
	  printf("*** %ld error(s) encountered while verifying data\n", (unsigned long) nErr);
	  bFailed = TRUE;
	  break;
	}
      }
      
      printf(" Reversed own address...\n");
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
	if (!(glob.bankTestMask & (0x1U << i))) {
	  continue;
	}
	
	if (bUseMask) {
	  displayMask(glob.bankWidth[i], mask, maskStr + 14);
	} else {
	  maskStr[0] = '\0';
	}
	
	printf("  Bank %ld%s\r", (unsigned long)i, maskStr);
	glob.pFpgaRegs[BANK_REG] = i & BANK_REG_BANKMASK;
	glob.pFpgaRegs[BANK_REG];
	width = glob.bankWidth[i];
	
	/* Write data to memory bank */
	remaining = glob.bankSize[i];
	offset = 0;
	while (remaining) {
	  chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  generateRevAddr(&glob, (uint32_t*)glob.pBufRevAddr, (uint32_t)(offset >> 2), chunk);
	  writeMemory(&glob, glob.pBufRevAddr, offset, chunk, dma);
	  remaining -= chunk;
	  offset += chunk;
	  spin(&glob, &spindex);
	}
        
	/* Read back data from memory bank and check */
	remaining = glob.bankSize[i];
	offset = 0;
	location = 0;
	nErr = 0;
	nRetry = 20;
	chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	while (remaining) {
	  unsigned long pos;
	  uint64_t tempLoc = location;
	  uint64_t errLoc = 0; /* init. only to prevent warning from compiler */
	  unsigned long errPos = 0; /* init. only to prevent warning from compiler */
	  BOOL bOk = TRUE;
          
	  readMemory(&glob, glob.pBufChk, offset, chunk, dma);
	  generateRevAddr(&glob, (uint32_t*)glob.pBufRevAddr, (uint32_t)(offset >> 2), chunk);
	  tempLoc = location;
	  for (pos = 0; pos < chunk; pos += width) {
	    maskExpectedWord(exp, glob.pBufRevAddr + pos, mask, width);
	    if (verifyWord(glob.pBufChk + pos, exp, act, mask, width)) {
	      nErr++;
	      displayError(tempLoc, exp, act, width, nErr);
	      if (bOk) {
		bOk = FALSE;
		errLoc = tempLoc;
		errPos = pos;
	      }
	    }
	    tempLoc++;
	  }
	  if (bOk || !bRetryReads) {
	    remaining -= chunk;
	    offset += chunk;
	    location += chunk;
	    nRetry = 20;
	    chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  } else {
	    if (0 == nRetry) {
	      printf("[Giving up on location 0x%llx]\n", errLoc);
	      remaining -= chunk;
	      offset += chunk;
	      location += chunk;
	      nRetry = 20;
	      chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	    } else {
	      printf("[Retrying read at location 0x%llx]\n", errLoc);
	      remaining -= errPos;
	      offset += errPos;
	      location = errLoc;
	      nRetry--;
	      chunk = width;
	    }
	  }
	  spin(&glob, &spindex);
	}
	if (nErr) {
	  printf("*** %ld error(s) encountered while verifying data\n", (unsigned long) nErr);
	  bFailed = TRUE;
	  break;
	}
      }
      
      printf(" Random data...\n");
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
	if (!(glob.bankTestMask & (0x1U << i))) {
	  continue;
	}
	if (bUseMask) {
	  displayMask(glob.bankWidth[i], mask, maskStr + 14);
	} else {
	  maskStr[0] = '\0';
	}
	printf("  Bank %ld%s\r", (unsigned long)i, maskStr);
	glob.pFpgaRegs[BANK_REG] = i & BANK_REG_BANKMASK;
	glob.pFpgaRegs[BANK_REG];
	width = glob.bankWidth[i];
	
	/* Write data to memory bank */
	remaining = glob.bankSize[i];
	offset = 0;
	seed(&glob.mtState, i + 1);
	while (remaining) {
	  chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  generateRandom(&glob.mtState, (uint32_t*)glob.pBufRandom, chunk);
	  writeMemory(&glob, glob.pBufRandom, offset, chunk, dma);
	  remaining -= chunk;
	  offset += chunk;
	  spin(&glob, &spindex);
	}
        
	/* Read back data from memory bank and check */
	remaining = glob.bankSize[i];
	offset = 0;
	location = 0;
	nErr = 0;
	nRetry = 20;
	chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	seed(&glob.mtState, i + 1);
	while (remaining) {
	  unsigned long pos;
	  uint64_t tempLoc = location;
	  uint64_t errLoc = 0; /* init. only to prevent warning from compiler */
	  unsigned long errPos = 0; /* init. only to prevent warning from compiler */
	  BOOL bOk = TRUE;
          
	  readMemory(&glob, glob.pBufChk, offset, chunk, dma);
	  generateRandom(&glob.mtState, (uint32_t*)glob.pBufRandom, chunk);
	  tempLoc = location;
	  for (pos = 0; pos < chunk; pos += width) {
	    maskExpectedWord(exp, glob.pBufRandom + pos, mask, width);
	    if (verifyWord(glob.pBufChk + pos, exp, act, mask, width)) {
	      nErr++;
	      displayError(tempLoc, exp, act, width, nErr);
	      if (bOk) {
		bOk = FALSE;
		errLoc = tempLoc;
		errPos = pos;
	      }
	    }
	    tempLoc++;
	  }
	  if (bOk || !bRetryReads) {
	    remaining -= chunk;
	    offset += chunk;
	    location += chunk;
	    nRetry = 20;
	    chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	  } else {
	    if (0 == nRetry) {
	      printf("[Giving up on location 0x%llx]\n", errLoc);
	      remaining -= chunk;
	      offset += chunk;
	      location += chunk;
	      nRetry = 20;
	      chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
	    } else {
	      printf("[Retrying read at location 0x%llx]\n", errLoc);
	      remaining -= errPos;
	      offset += errPos;
	      location = errLoc;
	      nRetry--;
	      chunk = width;
	    }
	  }
	  spin(&glob, &spindex);
	}
	if (nErr) {
	  printf("*** %ld error(s) encountered while verifying data\n", (unsigned long) nErr);
	  bFailed = TRUE;
	  break;
	}
      }
      if (bFailed) {
	break;
      }
    }
    if (bFailed) {
      break;
    }
  }
  
  /***********************************************************************
  ** 21. Here we measure the throughput attainable between the host and **
  **     each memory bank (unless disabled on the command-line).        **
  ***********************************************************************/
  
  if (bMeasurePerf) {
    for (dma = 0; dma < 2; dma++) {
      if (!dma && !bUsePio) {
	continue;
      }
      if (dma && !bUseDma) {
	continue;
      }
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
	const unsigned long chunk = GENERATE_CHUNK;
	float bytes, rate;
	struct timeval start, now, finish, delta;
	
	if (!(glob.bankTestMask & (0x1U << i))) {
	  continue;
	}
	
	printf("Measuring throughput for bank %ld using %s...\n",
	       (unsigned long) i, dma ? "DMA" : "programmed I/O");
	glob.pFpgaRegs[BANK_REG] = i & BANK_REG_BANKMASK;
	glob.pFpgaRegs[BANK_REG];
	
	start = get_time();
	finish = make_time(2, 0);
	finish = time_add(&finish, &start);
	bytes = 0.0f;
	do {
	  readMemory(&glob, glob.pBufChk, 0, chunk, dma);
	  bytes += (float) chunk;
	  now = get_time();
	} while (time_before(&now, &finish));
	delta = time_subtract(&now, &start);
	rate = bytes / time_to_float(&delta);
	printf(" SSRAM to host throughput = %.1fMB/s\n", rate / 1048576.0f);
	
	start = get_time();
	finish = make_time(2, 0);
	finish = time_add(&finish, &start);
	bytes = 0.0f;
	do {
	  writeMemory(&glob, glob.pBufChk, 0, chunk, dma);
	  bytes += (float) chunk;
	  now = get_time();
	} while (time_before(&now, &finish));
	delta = time_subtract(&now, &start);
	rate = bytes / time_to_float(&delta);
	printf(" Host to SSRAM throughput = %.1fMB/s\n", rate / 1048576.0f);
      }
    }
  }
  
  /***********************************************************************
  ** 22. Clean up.                                                      **
  ***********************************************************************/

 done:
  if (glob.pBuf5a != NULL) {
    free(glob.pBuf5a);
  }

  if (glob.pBufOwnAddr != NULL) {
    free(glob.pBufOwnAddr);
  }
  
  if (glob.pBufRevAddr != NULL) {
    free(glob.pBufRevAddr);
  }
  
  if (glob.pBufRandom != NULL) {
    free(glob.pBufRandom);
  }
  
  if (glob.pBufChk != NULL) {
    free(glob.pBufChk);
  }
  
  if (glob.hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(glob.hCard);
  }
  
  return ret;
}
