--
-- ddrcapture_reg.vhd - Capture register for DDR memory interfaces
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
library std;
use std.textio.all;
-- synopsys translate_on

library work;
use work.memif.all;

entity ddrcapture_reg is
    generic(
        family     : in  integer;
        width      : in  natural);
    port (
        k0         : in  std_logic;
        k1         : in  std_logic;
        d          : in  std_logic_vector(width - 1 downto 0);
        q0         : out std_logic_vector(width - 1 downto 0);
        q1         : out std_logic_vector(width - 1 downto 0));
end;

architecture mixed of ddrcapture_reg is

    -- synopsys translate_off
    function conv_string(
        val : in integer)
    return string is
        variable l : line;
        variable len : integer;
        variable s : string(1 to 64);
    begin
        write(l, val);
        len := l'length;
        read(l, s(1 to len));
        return s(1 to len);
    end;
    -- synopsys translate_on

    signal logic0 : std_logic;
    signal logic1 : std_logic;

    --
    -- XST synthesis constraints, required to prevent optimization of
    -- flip-flops that would otherwise prevent IOB packing.
    --
    attribute equivalent_register_removal : string;
    attribute equivalent_register_removal of q0 : signal is "no";
    attribute equivalent_register_removal of q1 : signal is "no";
    attribute iob : string;
    attribute iob of q0 : signal is "true";
    attribute iob of q1 : signal is "true";

begin

    logic0 <= '0';
    logic1 <= '1';

    capture0 : process(k0)
    begin
        if k0'event and k0 = '1' then
            q0 <= d;
        end if;
    end process;

    capture1 : process(k1)
    begin
        if k1'event and k1 = '1' then
            q1 <= d;
        end if;
    end process;

end mixed;
