--
-- null_mgt.vhd - This file should be included in a Virtex-4FX design
--                so that the workarounds for Virtex-4 MGT errata are
--                present in the bitstream.
--
-- (C) Copyright Alpha Data Parallel Systems Ltd. 2006-2007
--
-- Library code for:
--
--   o ADM-XRC-4FX
--
-- Instantiates a number of NULL_PAIR macros, one per MGT pair. Makes use of
-- core 'NULL_PAIR.NGC'.
--

library ieee;
use ieee.std_logic_1164.all;

entity null_mgt is
    port(
        grefclk          : in    std_logic;
        pcie_rx_p        : in    std_logic_vector(7 downto 0);
        pcie_rx_n        : in    std_logic_vector(7 downto 0);
        pcie_tx_p        : out   std_logic_vector(7 downto 0);
        pcie_tx_n        : out   std_logic_vector(7 downto 0);
        rear_rx_p        : in    std_logic_vector(3 downto 0);
        rear_rx_n        : in    std_logic_vector(3 downto 0);
        rear_tx_p        : out   std_logic_vector(3 downto 0);
        rear_tx_n        : out   std_logic_vector(3 downto 0);
        front_rx_p       : in    std_logic_vector(3 downto 0);
        front_rx_n       : in    std_logic_vector(3 downto 0);
        front_tx_p       : out   std_logic_vector(3 downto 0);
        front_tx_n       : out   std_logic_vector(3 downto 0));
end entity;

architecture struct of null_mgt is

    component NULL_PAIR
        port(
            GREFCLK_IN       : in    std_logic;   
            RX1N_IN          : in    std_logic_vector(1 downto 0);
            RX1P_IN          : in    std_logic_vector(1 downto 0);
            TX1N_OUT         : out   std_logic_vector(1 downto 0);
            TX1P_OUT         : out   std_logic_vector(1 downto 0));
    end component;

    attribute BOX_TYPE : string;
    attribute BOX_TYPE of NULL_PAIR : component is "user_black_box";

begin

    --
    -- PCI-E MGT lines
    --

    pcie_null_gen : for i in 0 to 3 generate
        pcie_inst : NULL_PAIR
      	    port map(
                GREFCLK_IN => grefclk,
                RX1N_IN    => pcie_rx_n((2 * i + 1) downto 2 * i),
                RX1P_IN    => pcie_rx_p((2 * i + 1) downto 2 * i),
                TX1N_OUT   => pcie_tx_n((2 * i + 1) downto 2 * i), 
                TX1P_OUT   => pcie_tx_p((2 * i + 1) downto 2 * i));
    end generate;

    --
    -- Rear MGT lines on RHS of device
    --

    rear_null_gen : for i in 0 to 1 generate
        rear_inst : NULL_PAIR
            port map(
                GREFCLK_IN => grefclk,
                RX1N_IN    => rear_rx_n((2 * i + 1) downto 2 * i),
                RX1P_IN    => rear_rx_p((2 * i + 1) downto 2 * i),
                TX1N_OUT   => rear_tx_n((2 * i + 1) downto 2 * i), 
                TX1P_OUT   => rear_tx_p((2 * i + 1) downto 2 * i));
    end generate;
	
    --
    -- Front (XRM) MGT lines on RHS of device
    --

    front_null_gen : for i in 0 to 1 generate
        front_inst : NULL_PAIR
            port map(
                GREFCLK_IN => grefclk,
                RX1N_IN    => front_rx_n((2*i+1) downto 2*i),
                RX1P_IN    => front_rx_p((2*i+1) downto 2*i),
                TX1N_OUT   => front_tx_n((2*i+1) downto 2*i), 
                TX1P_OUT   => front_tx_p((2*i+1) downto 2*i));
    end generate;
            
end struct;
