--
-- memory_banks-xrc4fx.vhd - Memory banks for ADM-XRC-4FX
--
-- (C) Copyright Alpha Data 2005-2007
--
-- SYNTHESIZABLE
--
-- Targets models:
--
--   o ADM-XRC-4FX
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

library work;
use work.memif.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity memory_banks is
    generic(
        addr_width    : in    natural;
        data_width    : in    natural;
        bank0         : in    memif_bank_t;
        bank1         : in    memif_bank_t;
        bank2         : in    memif_bank_t;
        bank3         : in    memif_bank_t;
        bank4         : in    memif_bank_t;
        bank5         : in    memif_bank_t;
        bank6         : in    memif_bank_t;
        bank7         : in    memif_bank_t;
        bank8         : in    memif_bank_t;
        bank9         : in    memif_bank_t;
        bank10        : in    memif_bank_t;
        bank11        : in    memif_bank_t;
        bank12        : in    memif_bank_t;
        bank13        : in    memif_bank_t;
        bank14        : in    memif_bank_t;
        bank15        : in    memif_bank_t;
        num_ramclk    : in    natural);
    port(
        rst           : in    std_logic;
        clk           : in    std_logic;
        mclk_i        : in    std_logic;
        refclk_i      : in    std_logic; -- 200MHz clock from oscillator
        sel_bank_1h   : in    std_logic_vector(15 downto 0);
        bank_reg      : in    std_logic_vector(3 downto 0);
        mode_reg      : in    std_logic_vector(511 downto 0);
        errata_reg    : in    std_logic_vector(0 downto 0);
        mem_ce        : in    std_logic;
        mem_term      : in    std_logic;
        mem_cw        : in    std_logic;
        mem_adv       : in    std_logic;
        mem_wr        : in    std_logic;
        mem_a         : in    std_logic_vector(addr_width - 1 downto 0);
        mem_d         : in    std_logic_vector(data_width - 1 downto 0);
        mem_be        : in    std_logic_vector(data_width / 8 - 1 downto 0);
        mem_q         : out   std_logic_vector(data_width - 1 downto 0);
        mem_wf        : out   std_logic;
        mem_wpf       : out   std_logic;
        mem_re        : out   std_logic;
        mem_rpe       : out   std_logic;
        locked        : out   std_logic_vector(7 downto 0);
        ra0           : out   std_logic_vector(bank0.ra_width - 1 downto 0);
        rc0           : inout std_logic_vector(bank0.rc_width - 1 downto 0);
        rd0           : inout std_logic_vector(bank0.rd_width - 1 downto 0);
        ra1           : out   std_logic_vector(bank1.ra_width - 1 downto 0);
        rc1           : inout std_logic_vector(bank1.rc_width - 1 downto 0);
        rd1           : inout std_logic_vector(bank1.rd_width - 1 downto 0);
        ra2           : out   std_logic_vector(bank2.ra_width - 1 downto 0);
        rc2           : inout std_logic_vector(bank2.rc_width - 1 downto 0);
        rd2           : inout std_logic_vector(bank2.rd_width - 1 downto 0);
        ra3           : out   std_logic_vector(bank3.ra_width - 1 downto 0);
        rc3           : inout std_logic_vector(bank3.rc_width - 1 downto 0);
        rd3           : inout std_logic_vector(bank3.rd_width - 1 downto 0);
        ra4           : out   std_logic_vector(bank4.ra_width - 1 downto 0);
        rc4           : inout std_logic_vector(bank4.rc_width - 1 downto 0);
        rd4           : inout std_logic_vector(bank4.rd_width - 1 downto 0);
        ra5           : out   std_logic_vector(bank5.ra_width - 1 downto 0);
        rc5           : inout std_logic_vector(bank5.rc_width - 1 downto 0);
        rd5           : inout std_logic_vector(bank5.rd_width - 1 downto 0);
        ra6           : out   std_logic_vector(bank6.ra_width - 1 downto 0);
        rc6           : inout std_logic_vector(bank6.rc_width - 1 downto 0);
        rd6           : inout std_logic_vector(bank6.rd_width - 1 downto 0);
        ra7           : out   std_logic_vector(bank7.ra_width - 1 downto 0);
        rc7           : inout std_logic_vector(bank7.rc_width - 1 downto 0);
        rd7           : inout std_logic_vector(bank7.rd_width - 1 downto 0);
        ra8           : out   std_logic_vector(bank8.ra_width - 1 downto 0);
        rc8           : inout std_logic_vector(bank8.rc_width - 1 downto 0);
        rd8           : inout std_logic_vector(bank8.rd_width - 1 downto 0);
        ra9           : out   std_logic_vector(bank9.ra_width - 1 downto 0);
        rc9           : inout std_logic_vector(bank9.rc_width - 1 downto 0);
        rd9           : inout std_logic_vector(bank9.rd_width - 1 downto 0);
        ra10          : out   std_logic_vector(bank10.ra_width - 1 downto 0);
        rc10          : inout std_logic_vector(bank10.rc_width - 1 downto 0);
        rd10          : inout std_logic_vector(bank10.rd_width - 1 downto 0);
        ra11          : out   std_logic_vector(bank11.ra_width - 1 downto 0);
        rc11          : inout std_logic_vector(bank11.rc_width - 1 downto 0);
        rd11          : inout std_logic_vector(bank11.rd_width - 1 downto 0);
        ra12          : out   std_logic_vector(bank12.ra_width - 1 downto 0);
        rc12          : inout std_logic_vector(bank12.rc_width - 1 downto 0);
        rd12          : inout std_logic_vector(bank12.rd_width - 1 downto 0);
        ra13          : out   std_logic_vector(bank13.ra_width - 1 downto 0);
        rc13          : inout std_logic_vector(bank13.rc_width - 1 downto 0);
        rd13          : inout std_logic_vector(bank13.rd_width - 1 downto 0);
        ra14          : out   std_logic_vector(bank14.ra_width - 1 downto 0);
        rc14          : inout std_logic_vector(bank14.rc_width - 1 downto 0);
        rd14          : inout std_logic_vector(bank14.rd_width - 1 downto 0);
        ra15          : out   std_logic_vector(bank15.ra_width - 1 downto 0);
        rc15          : inout std_logic_vector(bank15.rc_width - 1 downto 0);
        rd15          : inout std_logic_vector(bank15.rd_width - 1 downto 0);
        ramclki       : in    std_logic_vector(num_ramclk - 1 downto 0);
        ramclko       : out   std_logic_vector(num_ramclk - 1 downto 0));
end entity;

architecture mixed of memory_banks is

    constant mux_order : natural := 2;
    constant num_bank : natural := 4;

    constant port_width : natural := 128;
    constant port_be_width : natural := port_width / 8;
    constant port_addr_width : natural := addr_width - mux_order;

    constant ra_width : natural := num_bank * 19;
    constant rc_width : natural := num_bank * 22;
    constant rd_width : natural := num_bank * 32;

    signal logic0, logic1 : std_logic;

    signal memclk0 : std_logic;
    signal memclk45 : std_logic;
    signal memclk2x0 : std_logic;
    signal memclk2x90 : std_logic;
    signal memrst : std_logic;

    signal rep_be : std_logic_vector(port_be_width - 1 downto 0);
    signal rep_d : std_logic_vector(port_width - 1 downto 0);
    signal rep_valid : std_logic;
    signal rep_final : std_logic;

    constant tag_width : natural := 2;
    subtype tag_t is std_logic_vector(tag_width - 1 downto 0);
    type tag_vector_t is array(natural range <>) of tag_t;

    signal tag_base : std_logic_vector(tag_width - 1 downto 0);
    signal tag_mask : std_logic_vector(tag_width - 1 downto 0);

    type port_pq_t is array(0 to num_bank - 1) of std_logic_vector(port_width - 1 downto 0);
    signal port_pce : std_logic_vector(num_bank - 1 downto 0);
    signal port_pcw : std_logic_vector(num_bank - 1 downto 0);
    signal port_pterm : std_logic_vector(num_bank - 1 downto 0);
    signal port_padv : std_logic_vector(num_bank - 1 downto 0);
    signal port_pwr : std_logic_vector(num_bank - 1 downto 0);
    signal port_pq : port_pq_t;
    signal port_pwf : std_logic_vector(num_bank - 1 downto 0);
    signal port_pwpf : std_logic_vector(num_bank - 1 downto 0);
    signal port_pre : std_logic_vector(num_bank - 1 downto 0);
    signal port_prpe : std_logic_vector(num_bank - 1 downto 0);

    signal port_plast : std_logic;
    signal port_pq_muxed : std_logic_vector(port_width - 1 downto 0);

    type port_sa_t is array(0 to num_bank - 1) of std_logic_vector(port_addr_width - 1 downto 0);
    type port_sd_t is array(0 to num_bank - 1) of std_logic_vector(port_width - 1 downto 0);
    type port_sbe_t is array(0 to num_bank - 1) of std_logic_vector(port_be_width - 1 downto 0);
    type port_stag_t is array(0 to num_bank - 1) of std_logic_vector(tag_width - 1 downto 0);
    signal port_sce : std_logic_vector(num_bank - 1 downto 0);
    signal port_sw : std_logic_vector(num_bank - 1 downto 0);
    signal port_stag : port_stag_t;
    signal port_sa : port_sa_t;
    signal port_sd : port_sd_t;
    signal port_sbe : port_sbe_t;
    signal port_sq : port_sd_t;
    signal port_sqtag : port_stag_t;
    signal port_svalid : std_logic_vector(num_bank - 1 downto 0);
    signal port_sready : std_logic_vector(num_bank - 1 downto 0);

    component memclk
        port(
            mclk              : in  std_logic;
            rst               : in  std_logic;
            refclk200         : in  std_logic;
            clk0              : out std_logic;
            clk45             : out std_logic;
            clk2x0            : out std_logic;
            clk2x90           : out std_logic;
            locked            : out std_logic_vector(1 downto 0);
            memrst            : out std_logic);
    end component;

begin

    logic0 <= '0';
    logic1 <= '1';

    mem_wf <= OR_reduce(sel_bank_1h(num_bank - 1 downto 0) and port_pwf);
    mem_wpf <= OR_reduce(sel_bank_1h(num_bank - 1 downto 0) and port_pwpf);
    mem_re <= OR_reduce(sel_bank_1h(num_bank - 1 downto 0) and port_pre);
    mem_rpe <= OR_reduce(sel_bank_1h(num_bank - 1 downto 0) and port_prpe);

    locked(7 downto 6) <= (others => '0');

    --
    -- Instantiate logic to generate memory clocks
    --

    memclk_0 : memclk
        port map(
            mclk => mclk_i,
            rst => rst,
            refclk200 => refclk_i,
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            locked => locked(1 downto 0),
            memrst => memrst);

    --
    -- Instantiate outbound data replicator
    --

    port_repl_0 : port_repl
        generic map(
            order => mux_order,
            in_width => 32,
            out_width => port_width,
            partial => true)
        port map(
            rst => memrst,
            clk => clk,
            init => mem_ce,
            addr => mem_a(mux_order - 1 downto 0),
            wr => mem_wr,
            last => mem_term,
            din => mem_d,
            bein => mem_be,
            dout => rep_d,
            beout => rep_be,
            valid => rep_valid,
            final => rep_final);

    --
    -- Select data from selected bank
    --

    gen_port_pq_muxed : process(
        port_pq,
        sel_bank_1h)
        variable x : std_logic_vector(port_width - 1 downto 0);
    begin
        x := (others => '0');
        for i in 0 to num_bank - 1 loop
            if sel_bank_1h(i) = '1' then
                x := x or port_pq(i);
            end if;
        end loop;
        port_pq_muxed <= x;
    end process;

    --
    -- Instantiate inbound data multiplexor
    --

    port_mux_0 : port_mux
        generic map(
            order => mux_order,
            in_width => port_width,
            out_width => 32)
        port map(
            rst => memrst,
            clk => clk,
            init => mem_ce,
            addr => mem_a(mux_order - 1 downto 0),
            adv => mem_adv,
            din => port_pq_muxed,
            dout => mem_q,
            last => port_plast);

    --
    -- Tags used by local bus to memory interface are "00" and "01", leaving
    -- "10" and "11" spare.
    --

    tag_base <= EXT("00", tag_width);
    tag_mask <= EXT("01", tag_width);

    --
    -- Instantiate 'async_port' components for DDR-II SDRAM memory banks.
    --
    -- These instances decouple the local bus clock domain from the
    -- memory interface clock domain.
    --

    gen_async_ports : for i in 0 to num_bank - 1 generate
        port_pce(i) <= sel_bank_1h(i) and mem_ce;
        port_pcw(i) <= mem_cw;
        port_pterm(i) <= sel_bank_1h(i) and rep_final;
        port_padv(i) <= sel_bank_1h(i) and port_plast and mem_adv;
        port_pwr(i) <= sel_bank_1h(i) and rep_valid;

        async_port_0 : async_port
            generic map(
                family => memif_model_to_family(memif_model_admxrc4fx),
                order => 5,
                iwpfl => 3,
                orpel => 1,
                owpfl => 22,
                addr_width => port_addr_width,
                data_width => port_width,
                tag_width => 2)
            port map(
                rst => memrst,
                pclk => clk,
                psr => logic0,
                pce => port_pce(i),
                pcw => port_pcw(i),
                pterm => port_pterm(i),
                padv => port_padv(i),
                pwr => port_pwr(i),
                pa => mem_a(addr_width - 1 downto mux_order),
                pd => rep_d,
                pbe => rep_be,
                pq => port_pq(i),
                pwf => port_pwf(i),
                pwpf => port_pwpf(i),
                pre => port_pre(i),
                prpe => port_prpe(i),
                sclk => memclk0,
                sce => port_sce(i),
                sw => port_sw(i),
                stag => port_stag(i),
                sa => port_sa(i),
                sd => port_sd(i),
                sbe => port_sbe(i),
                sq => port_sq(i),
                sqtag => port_sqtag(i),
                svalid => port_svalid(i),
                sready => port_sready(i),
                tag_base => tag_base,
                tag_mask => tag_mask);
    end generate;

    dram_port_0 : ddr2sdram_port
        generic map(
            model => memif_model_admxrc4fx,
            bank_index => 0,
            ra_width => 19,
            rc_width => 22,
            rd_width => 32,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => 2,
            -- All timings are in 'clk0' cycles
            t_refresh => 350, -- Ok for 'clk0' >= 100MHz, industrial parts (3.9us refresh)
            t_mrd => 3,       -- Use value 3 for safety
            t_dllr => 150,    -- Datasheets require 200 clk2x0 cycles / 100 clk0 cycles
            -- Timings for 266MHz 'clk2x0', 133MHz 'clk0'
            t_rp => 2,
            t_rfc => 27,
            t_act => 2,
            t_wtr => 2,
            t_rtw => 3,
            t_rtp => 2,
            t_wtp => 5,
            t_ras => 6)
        port map(
            rst => memrst,
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            sr => logic0,
            ce => port_sce(0),
            w => port_sw(0),
            tag => port_stag(0),
            a => port_sa(0),
            d => port_sd(0),
            be => port_sbe(0),
            q => port_sq(0),
            qtag => port_sqtag(0),
            ready => port_sready(0),
            valid => port_svalid(0),
            row => mode_reg(32 * 0 + 3 downto 32 * 0 + 2),
            col => mode_reg(32 * 0 + 5 downto 32 * 0 + 4),
            bank => mode_reg(32 * 0 + 7 downto 32 * 0 + 6),
            pbank => mode_reg(32 * 0 + 9 downto 32 * 0 + 8),
            trained => locked(2),
            ra => ra0,
            rc => rc0,
            rd => rd0);

    dram_port_1 : ddr2sdram_port
        generic map(
            model => memif_model_admxrc4fx,
            bank_index => 1,
            ra_width => 19,
            rc_width => 22,
            rd_width => 32,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => 2,
            -- All timings are in 'clk0' cycles
            t_refresh => 350, -- Ok for 'clk0' >= 100MHz, industrial parts (3.9us refresh)
            t_mrd => 3,       -- Use value 3 for safety
            t_dllr => 150,    -- Datasheets require 200 clk2x0 cycles / 100 clk0 cycles
            -- Timings for 266MHz 'clk2x0', 133MHz 'clk0'
            t_rp => 2,
            t_rfc => 27,
            t_act => 2,
            t_wtr => 2,
            t_rtw => 3,
            t_rtp => 2,
            t_wtp => 5,
            t_ras => 6)
        port map(
            rst => memrst,
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            sr => logic0,
            ce => port_sce(1),
            w => port_sw(1),
            tag => port_stag(1),
            a => port_sa(1),
            d => port_sd(1),
            be => port_sbe(1),
            q => port_sq(1),
            qtag => port_sqtag(1),
            ready => port_sready(1),
            valid => port_svalid(1),
            row => mode_reg(32 * 1 + 3 downto 32 * 1 + 2),
            col => mode_reg(32 * 1 + 5 downto 32 * 1 + 4),
            bank => mode_reg(32 * 1 + 7 downto 32 * 1 + 6),
            pbank => mode_reg(32 * 1 + 9 downto 32 * 1 + 8),
            trained => locked(3),
            ra => ra1,
            rc => rc1,
            rd => rd1);

    dram_port_2 : ddr2sdram_port
        generic map(
            model => memif_model_admxrc4fx,
            bank_index => 2,
            ra_width => 19,
            rc_width => 22,
            rd_width => 32,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => 2,
            -- All timings are in 'clk0' cycles
            t_refresh => 350, -- Ok for 'clk0' >= 100MHz, industrial parts (3.9us refresh)
            t_mrd => 3,       -- Use value 3 for safety
            t_dllr => 150,    -- Datasheets require 200 clk2x0 cycles / 100 clk0 cycles
            -- Timings for 266MHz 'clk2x0', 133MHz 'clk0'
            t_rp => 2,
            t_rfc => 27,
            t_act => 2,
            t_wtr => 2,
            t_rtw => 3,
            t_rtp => 2,
            t_wtp => 5,
            t_ras => 6)
        port map(
            rst => memrst,
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            sr => logic0,
            ce => port_sce(2),
            w => port_sw(2),
            tag => port_stag(2),
            a => port_sa(2),
            d => port_sd(2),
            be => port_sbe(2),
            q => port_sq(2),
            qtag => port_sqtag(2),
            ready => port_sready(2),
            valid => port_svalid(2),
            row => mode_reg(32 * 2 + 3 downto 32 * 2 + 2),
            col => mode_reg(32 * 2 + 5 downto 32 * 2 + 4),
            bank => mode_reg(32 * 2 + 7 downto 32 * 2 + 6),
            pbank => mode_reg(32 * 2 + 9 downto 32 * 2 + 8),
            trained => locked(4),
            ra => ra2,
            rc => rc2,
            rd => rd2);

    dram_port_3 : ddr2sdram_port
        generic map(
            model => memif_model_admxrc4fx,
            bank_index => 3,
            ra_width => 19,
            rc_width => 22,
            rd_width => 32,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => 2,
            -- All timings are in 'clk0' cycles
            t_refresh => 350, -- Ok for 'clk0' >= 100MHz, industrial parts (3.9us refresh)
            t_mrd => 3,       -- Use value 3 for safety
            t_dllr => 150,    -- Datasheets require 200 clk2x0 cycles / 100 clk0 cycles
            -- Timings for 266MHz 'clk2x0', 133MHz 'clk0'
            t_rp => 2,
            t_rfc => 27,
            t_act => 2,
            t_wtr => 2,
            t_rtw => 3,
            t_rtp => 2,
            t_wtp => 5,
            t_ras => 6)
        port map(
            rst => memrst,
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            sr => logic0,
            ce => port_sce(3),
            w => port_sw(3),
            tag => port_stag(3),
            a => port_sa(3),
            d => port_sd(3),
            be => port_sbe(3),
            q => port_sq(3),
            qtag => port_sqtag(3),
            ready => port_sready(3),
            valid => port_svalid(3),
            row => mode_reg(32 * 3 + 3 downto 32 * 3 + 2),
            col => mode_reg(32 * 3 + 5 downto 32 * 3 + 4),
            bank => mode_reg(32 * 3 + 7 downto 32 * 3 + 6),
            pbank => mode_reg(32 * 3 + 9 downto 32 * 3 + 8),
            trained => locked(5),
            ra => ra3,
            rc => rc3,
            rd => rd3);

    --
    -- Banks 4-15 are not present/used.
    --

    ra4 <= (others => 'Z');
    rc4 <= (others => 'Z');
    rd4 <= (others => 'Z');
    ra5 <= (others => 'Z');
    rc5 <= (others => 'Z');
    rd5 <= (others => 'Z');
    ra6 <= (others => 'Z');
    rc6 <= (others => 'Z');
    rd6 <= (others => 'Z');
    ra7 <= (others => 'Z');
    rc7 <= (others => 'Z');
    rd7 <= (others => 'Z');
    ra8 <= (others => 'Z');
    rc8 <= (others => 'Z');
    rd8 <= (others => 'Z');
    ra9 <= (others => 'Z');
    rc9 <= (others => 'Z');
    rd9 <= (others => 'Z');
    ra10 <= (others => 'Z');
    rc10 <= (others => 'Z');
    rd10 <= (others => 'Z');
    ra11 <= (others => 'Z');
    rc11 <= (others => 'Z');
    rd11 <= (others => 'Z');
    ra12 <= (others => 'Z');
    rc12 <= (others => 'Z');
    rd12 <= (others => 'Z');
    ra13 <= (others => 'Z');
    rc13 <= (others => 'Z');
    rd13 <= (others => 'Z');
    ra14 <= (others => 'Z');
    rc14 <= (others => 'Z');
    rd14 <= (others => 'Z');
    ra15 <= (others => 'Z');
    rc15 <= (others => 'Z');
    rd15 <= (others => 'Z');

    --
    -- No memory clocks to generate (clocks generated for each port)
    --

    ramclko <= (others => 'Z');

end architecture;
